/*
 * X11 GUI for slide puzzle game
 * Copyright © 2022-2023 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include <getopt.h>

#include <Xm/XmAll.h>

#include "help.h"
#include "xtra.h"
#include "motif.h"
#include "ewmhicon.h"
#include "motifopt.h"
#include "game.h"
#include "version.h"

#define TIMER_UPDATE_MS 33

#define SPLIT_NUMERATOR    75
#define SPLIT_DENOMINATOR 100

#define MIN(a, b) ((a) < (b) ? (a) : (b))

#define PROGNAME "rrace"
static const char *progname = PROGNAME;

static char * const default_resources[] = {
	"*title: RRace",
	"*game.height: 371", // 365 + 2*shadowThickness
	"*game.width: 498",

	"*game.XmFrame.shadowThickness: 3",
	"*game.XmFrame.shadowType: shadow_in",
	"*goalArea.leftOffset: 1",

	"*gameNew.accelerator: Ctrl<Key>N",
	"*gameNew.acceleratorText: Ctrl+N",
	"*gameExit.accelerator: Ctrl<Key>Q",
	"*gameExit.acceleratorText: Ctrl+Q",

	"*aboutDialog*pixmapTextPadding: 10",

	NULL
};

static void print_version(void)
{
	version_print_head("rrace-motif", stdout);
	puts("License GPLv3+: GNU GPL version 3 or any later version");
	puts("This is free software: you are free to change and redistribute it.");
	puts("There is NO WARRANTY, to the extent permitted by law.");
}

static void print_usage(FILE *f)
{
	fprintf(f, "Usage: %s [options]\n", progname);
	if (f != stdout)
		fprintf(f, "Try %s --help for more information.\n", progname);
}

static void print_help(const struct option *lopts)
{
	struct lopt_help help = {0};
	const struct option *opt;

	print_usage(stdout);

	putchar('\n');
	puts("Options:");
	for (opt = lopts; opt->name; opt++) {
		if (!lopt_get_help(opt, &help))
			continue;
		help_print_option(opt, help.arg, help.desc, 20);
	}
	putchar('\n');

	printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
}

static Widget early_setup(XtAppContext *app, int argc, char **argv)
{
	Widget shell;
	int opt;

	XTRA_PACKED_LOPTS(lopts);

	/* Check for --help/--version early (before X connection) */
	opterr = 0;
	while ((opt = getopt_long_only(argc, argv, "", lopts, NULL)) != -1) {
		switch (opt) {
		case LOPT_VERSION:
			print_version();
			exit(EXIT_SUCCESS);
		case LOPT_HELP:
			print_help(lopts);
			exit(EXIT_SUCCESS);
		}
	}

	shell = XtOpenApplication(app, PACKAGE_TARNAME, NULL, 0,
	                          &argc, argv, (String *)default_resources,
	                          sessionShellWidgetClass, NULL, 0);

	opterr = optind = 1;
	while ((opt = getopt_long_only(argc, argv, "", lopts, NULL)) != -1) {
		switch (opt) {
		default:
			print_usage(stderr);
			exit(EXIT_FAILURE);
		}
	}

	if (argv[optind]) {
		fprintf(stderr, "%s: excess command-line arguments.\n",
		                progname);
		print_usage(stderr);
		exit(EXIT_FAILURE);
	}

	return shell;
}

static void timer_tick(void *data, XtIntervalId *id)
{
	struct app_state *state = data;
	XtAppContext app;

	if (state->board.x > 4) {
		/* Game is over */
		state->flags &= ~FLAG_TIMER_RUNNING;
		return;
	}

	ui_timer_update(state, game_elapsed(&state->board));
	app = XtWidgetToApplicationContext(state->game);
	XtAppAddTimeOut(app, TIMER_UPDATE_MS, timer_tick, state);
}

static void do_input_move(struct app_state *state, int x, int y)
{
	uint_fast32_t mask;

	if ((mask = game_do_move(&state->board, x, y)) != 0) {
		if (game_check_goal(&state->board) == 0) {
			ui_timer_update(state, game_finish(&state->board));
			mask |= ~GOAL_MASK;
		}

		x11_redraw_game(state, mask);
	}
}

static void set_view_goal(struct app_state *state, int view_goal)
{
	state->flags &= ~FLAG_VIEW_GOAL_ON_GAME;
	state->flags |= view_goal ? FLAG_VIEW_GOAL_ON_GAME : 0;

	x11_redraw_game(state, game_check_goal(&state->board));
}

static void game_input(Widget w, void *data, void *cb_data)
{
	XmDrawingAreaCallbackStruct *cbs = cb_data;
	XButtonEvent *b = &cbs->event->xbutton;
	struct app_state *state = data;

	switch (cbs->event->type) {
	case ButtonPress:
		switch (b->button) {
		case Button1:
			if (b->state & Button3Mask)
				break;

			do_input_move(state, b->x / state->game_tile_sz,
			                     b->y / state->game_tile_sz);
			break;
		case Button3:
			set_view_goal(state, 1);
			break;
		}
		break;
	case ButtonRelease:
		switch (b->button) {
		case Button3:
			set_view_goal(state, 0);
			break;
		}
	}
}

static struct app_state state;

static Widget get_shell(Widget w)
{
	Widget shell;

	for (shell = w; !XtIsWMShell(shell);)
		shell = XtParent(shell);

	return shell;
}

static void proc_exit(Widget w, XEvent *e, String *argv, Cardinal *argc)
{
	XtAppSetExitFlag(XtWidgetToApplicationContext(w));
}

static void proc_new_game(Widget w, XEvent *e, String *argv, Cardinal *argc)
{
	XtAppContext app;

	game_reset(&state.board);

	state.flags &= ~FLAG_VIEW_GOAL_ON_GAME;
	x11_redraw_goal(&state, -1, get_shell(w));
	x11_redraw_game(&state, -1);

	game_begin(&state.board);
	if (!(state.flags & FLAG_TIMER_RUNNING)) {
		state.flags |= FLAG_TIMER_RUNNING;
		timer_tick(&state, 0);
	}
}

static void proc_about(Widget w, XEvent *e, String *argv, Cardinal *argc)
{
	ui_show_about(&state, get_shell(w));
}

static void proc_resize(Widget form, XEvent *e, String *argv, Cardinal *argc)
{
	Widget game = XtParent(state.game);
	Widget goal = XtParent(state.goal);
	Dimension w, h, gamesz, gameborder, goalsz, goalborder;
	int x, y, gap;

	XtVaGetValues(form, XmNwidth, &w, XmNheight, &h, (char *)NULL);
	XtVaGetValues(game, XmNshadowThickness, &gameborder, (char *)NULL);
	XtVaGetValues(goal, XmNshadowThickness, &goalborder,
	                    XmNleftOffset, &gap,
	                    (char *)NULL);

	gamesz = MIN(h, w * SPLIT_NUMERATOR / SPLIT_DENOMINATOR);
	state.game_tile_sz = (gamesz - 2*gameborder) / 5;
	gamesz = 5*state.game_tile_sz + 2*gameborder;

	goalsz = MIN(gamesz*3/5, w - gamesz - gap);
	state.goal_tile_sz = (goalsz - 2*goalborder) / 3;
	goalsz = 3*state.goal_tile_sz + 2*goalborder;

	x = (w - gamesz - goalsz - gap) / 2;
	if (x < 2) x = 0;

	y = (h - gamesz) / 2;
	if (y < 3) y = 0;

	XtVaSetValues(game, XmNleftOffset, x, XmNtopOffset, y, (char *)NULL);
	XtVaSetValues(game, XmNwidth, gamesz, XmNheight, gamesz, (char *)NULL);
	XtVaSetValues(goal, XmNwidth, goalsz, XmNheight, goalsz, (char *)NULL);
}

static const XtActionsRec app_actions[] = {
	{ "gameNew", proc_new_game },
	{ "gameExit", proc_exit },
	{ "helpAbout", proc_about },

	{ "ResizeGameArea", proc_resize }
};

static XtAppContext app_initialize(int argc, char **argv)
{
	XtAppContext app;
	Widget shell;
	int i;

	if (argc > 0)
		progname = argv[0];

	shell = early_setup(&app, argc, argv);
	XtAppAddActions(app, (void *)app_actions, XtNumber(app_actions));
	ui_initialize(&state, shell);
	x11_initialize(&state, shell);

	XtAddCallback(state.game, XmNinputCallback, game_input, &state);

	/* Begin with the game in winning state */
	game_reset(&state.board);
	for (i = 0; i < 3; i++)
		state.board.game[i] = state.board.goal[i] << GOAL_SHIFT;
	game_finish(&state.board);

	state.flags = ewmh_probe_wm_icon(shell);
	XtRealizeWidget(shell);

	x11_redraw_goal(&state, 0, shell);

	return app;
}

int main(int argc, char **argv)
{
	setlocale(LC_ALL, "");
	XtAppMainLoop(app_initialize(argc, argv));
	return 0;
}
